#include "AUMIDIEffectBase.h"

class Testau : public ausdk::AUMIDIEffectBase {
public:
  explicit Testau(AudioComponentInstance ci) : AUMIDIEffectBase{ci} {}

  bool StreamFormatWritable(AudioUnitScope scope,
                            AudioUnitElement element) override {
    return true;
  }

  bool CanScheduleParameters() const override { return false; }

  OSStatus HandleNoteOn(UInt8 /*inChannel*/, UInt8 /*inNoteNumber*/,
                        UInt8 /*inVelocity*/,
                        UInt32 /*inStartFrame*/) override {
    if (midiCallback.midiOutputCallback == nullptr ||
        midiCallback.userData == nullptr)
      return noErr;

    std::aligned_storage_t<(1 << 16)> storage;
    static_assert(sizeof(storage) == 1 << 16);
    auto &list = *reinterpret_cast<MIDIPacketList *>(&storage);

    auto *packet = MIDIPacketListInit(&list);

    for (auto i = 0; i < 256; ++i) {
      Byte data[]{(Byte)(0xe0 | (i % 16)), 0, 0};
      packet = MIDIPacketListAdd(&list, sizeof(storage), packet, 0,
                                 std::size(data), data);
    }

    midiCallback.midiOutputCallback(midiCallback.userData, &lastTimeStamp, 0,
                                    &list);

    return noErr;
  }

  OSStatus GetPropertyInfo(AudioUnitPropertyID inID, AudioUnitScope inScope,
                           AudioUnitElement inElement, UInt32 &outDataSize,
                           bool &outWritable) override {
    if (inScope == kAudioUnitScope_Global &&
        inID == kAudioUnitProperty_MIDIOutputCallbackInfo) {
      outDataSize = sizeof(CFArrayRef);
      outWritable = false;
      return noErr;
    }

    return AUMIDIEffectBase::GetPropertyInfo(inID, inScope, inElement,
                                             outDataSize, outWritable);
  }

  OSStatus GetProperty(AudioUnitPropertyID inID, AudioUnitScope inScope,
                       AudioUnitElement inElement, void *outData) override {
    if (inScope == kAudioUnitScope_Global &&
        inID == kAudioUnitProperty_MIDIOutputCallbackInfo) {
      const void *strs[]{
          reinterpret_cast<const void *>(CFSTR("MIDI Callback"))};
      auto callbackArray =
          CFArrayCreate(nullptr, strs, std::size(strs), &kCFTypeArrayCallBacks);
      *static_cast<CFArrayRef *>(outData) = callbackArray;
      return noErr;
    }

    return AUMIDIEffectBase::GetProperty(inID, inScope, inElement, outData);
  }

  OSStatus SetProperty(AudioUnitPropertyID inID, AudioUnitScope inScope,
                       AudioUnitElement inElement, const void *inData,
                       UInt32 inDataSize) override {
    if (inScope == kAudioUnitScope_Global &&
        inID == kAudioUnitProperty_MIDIOutputCallback) {

      if (inDataSize < sizeof(AUMIDIOutputCallbackStruct))
        return kAudioUnitErr_InvalidPropertyValue;

      if (auto *callbackStruct =
              static_cast<const AUMIDIOutputCallbackStruct *>(inData))
        midiCallback = *callbackStruct;

      return noErr;
    }

    return AUMIDIEffectBase::SetProperty(inID, inScope, inElement, inData,
                                         inDataSize);
  }

  OSStatus Render(AudioUnitRenderActionFlags &,
                  const AudioTimeStamp &inTimeStamp, UInt32) override {
    lastTimeStamp = inTimeStamp;
    return noErr;
  }

  AudioTimeStamp lastTimeStamp = {};
  AUMIDIOutputCallbackStruct midiCallback = {};
};

AUSDK_COMPONENT_ENTRY(ausdk::AUMIDIEffectFactory, Testau)
